/* Copyright (c) 2013 APDM
*
* longitudinal_plot.js
*
* Author: Jess VanDerwalker <jess@apdm.com>
*/

// JSHint ignores
/* global utils:false */

function longitudinalPlot() {
    // Physical dimentions
    var margin = { left:60, top:20, right:20, bottom:60 };
    var height = 400;
    var width = 400;
    var circleRadius = 6;

    // Scales and axis
    var minY = null;            // Minimum possible Y-value for axis range
    var xScale = d3.time.scale();
    var yScale = d3.scale.linear();
    var xAxis = d3.svg.axis().scale(xScale).orient("bottom")
        .ticks(2);
    var yAxis = d3.svg.axis().scale(yScale).orient("left")
        .ticks(2);

    // Labels
    var title = "";
    var xAxisLabel = "";
    var yAxisLabel = "";

    // Accessor functions for data
    var xValue = function(d) { return d[0]; };
    var yValue = function(d) { return d[1]; };
    var extraValues = function(d) { return d[2]; };

    // Normative values
    var normStdDev = null;
    var normMean = null;

    // Zoom behavior
    var zoomEnabled = false;
    var zoom = null;

    // Tooltips
    var tooltipsEnabled = false;
    var tooltipHtml = null;

    // On-click handler for points
    var pointOnClickHandler = null;

    function longPlot(selection) {
        selection.each(function(metricData) {
            if (!metricData || !metricData.length) {
                console.warn("No data provided to longitudinal plot");
                return;
            }

            // Format the data into a nice array
            metricData = metricData.map(function(d, i) {
                return [ xValue.call(metricData, d, i),
                         yValue.call(metricData, d, i),
                         extraValues.call(metricData, d, i) ];
            });
            
            // Set ranges and domains
            var dates = metricData.map(function(d) { return d[0]; });
            // If there is only one sample, then we need to pad the
            // date range so the value is not stuck to the left
            // edge. If there are more than two, add some padding to
            // either side to get the points off the edges of the
            // plot.
            if (dates && dates.length === 1) {
                var newDates = [];
                newDates[0] = new Date(dates[0].getTime());
                newDates[1] = new Date(dates[0].getTime());
                newDates[0].setHours(newDates[0].getHours() - 1);
                newDates[1].setHours(newDates[1].getHours() + 1);
                dates = newDates;
            } else {
                dates = d3.extent(dates);
                var range = dates[1].getTime() - dates[0].getTime();
                var datePad = range * 0.10; // 10% padding
                dates[0] = new Date(dates[0].getTime() - datePad);
                dates[1] = new Date(dates[1].getTime() + datePad);
            }
            xScale.domain(dates).nice()
                .range([0, width - margin.left - margin.right]);

            var yValues = metricData.map(function(d) { return d[1]; });
            // Add the mean and stdDev if they exist
            if (normMean && normStdDev) {
                yValues = yValues.concat([normMean, normMean + (2 * normStdDev),
                                          normMean - (2 * normStdDev)]);
            } else if (normMean) {
                yValues.push(normMean);
            }
            var yRange = d3.extent(yValues);
            if (minY !== null) {
                yRange[0] = minY;
            }
            yScale.domain(yRange).nice()
                .range([height - margin.top - margin.bottom,
                        circleRadius]);

            // Tooltip to be used if they have been enabled
            var tooltip = d3.select("body").append("div")
				.attr("class", "longitudinal_tooltip")
				.text("");

            var plotSvg = d3.select(this).selectAll("svg").data([metricData]);
            var plotG = plotSvg.enter().append("svg").append("g");
            plotG.append("defs").append("clipPath").attr("id", "clip-box")
                .append("rect");
            
            // Axis and label containers
            plotG.append("g").attr("class", "x axis");
            plotG.append("g").attr("class", "y axis");
            plotG.append("text").attr("class", "x label");
            plotG.append("text").attr("class", "y label");
            plotG.append("text").attr("class", "title label");

            // Only create the norm line and range if provided
            if (normMean) {
                plotG.append("line").attr("class", "norm_line");
                if (normStdDev) {
                    plotG.append("rect").attr("class", "norm_range_rect");
                }
            }

            plotG.append("rect").attr("class", "pointer_mask");

            // Update
            plotSvg.attr("width", width)
                .attr("height", height);
            
            plotG = plotSvg.select("g")
                .attr("transform", "translate(" + margin.left +
                      "," + margin.top + ")");

            plotG.selectAll("rect")
                .attr("width", width - margin.left - margin.right)
                .attr("height", height - margin.top - margin.bottom);

            plotG.select(".x.axis")
                .attr("transform", "translate(0," + yScale.range()[0] + ")")
                .call(xAxis);
            plotG.select(".y.axis")
                .attr("transform", "translate(" + xScale.range()[0] + ",0)")
                .call(yAxis);

            var xLabel = plotG.select(".x.label")
                .attr("text-anchor", "middle")
                .attr("x", (width - margin.left - margin.right)/2)
                .attr("y", (height))
                .text(xAxisLabel);

            var box = xLabel.node().getBBox();
            var hOffset = (height - margin.top - margin.bottom) +
                xAxis.tickPadding() + xAxis.tickSize() + box.height + (box.height / 2);
            xLabel.attr("y", hOffset);

            plotG.select(".y.label")
                .attr("text-anchor", "middle")
                .attr("x", (margin.top + margin.bottom - height)/2)
                .attr("y", -35)
                .attr("transform", "rotate(-90)")
                .text(yAxisLabel);

            plotG.select(".title.label")
                .attr("text-anchor", "middle")
                .attr("x", (width - margin.left - margin.right)/2)
                .attr("y", -10)
                .text(title);

            // Create the symbols on the plot
            var symbols = plotG.selectAll(".point")
                .data(metricData)
                .enter()
                .append("circle").attr("r", circleRadius)
                .attr("class", "point")
                .attr("clip-path", "url(#clip-box)")
                .attr("fill", "blue")
                .attr("cx", function(d) { return x(d); })
                .attr("cy", function(d) { return y(d); });

            // Set up tooltips if enabled
            if (tooltipsEnabled === true) {
                var tooltipLabel;
                    if (tooltipHtml === null) {
                        tooltipLabel = function(d) {
                            tooltip.html("<strong>x: </strong>" + d[0] +
                                         "<br><strong>y: </strong>" + d[1]);
                            if (event !== undefined) {
                                tooltip.style("top", (event.pageY - 10) + "px")
                                    .style("left", (event.pageX + 10) + "px");
                            }
                        };
                    } else {
                        tooltipLabel = function(d) {
                            tooltip.html(tooltipHtml(d));
                            if (event !== undefined) {
                                tooltip.style("top", (event.pageY - 10) + "px")
                                    .style("left", (event.pageX + 10) + "px");
                            }
                        };
                    }
                symbols.on("mouseover", function () {
                    return tooltip.style("visibility", "visible");
                }).on("mousemove", function (d) {
                    tooltipLabel(d);
                }).on("mouseout", function () {
                    tooltip.style("visibility", "hidden");
                });
            }
            
            // Add an on click handler if supplied
            if (pointOnClickHandler !== null) {
                symbols.on("click", pointOnClickHandler);
            }

            // Line for normal mean and rect for normal mean
            // range. This is currently 2 * normal standard deviation
            plotG.select(".norm_range_rect")
                .attr("x", 0)
                .attr("width", width - margin.left - margin.right)
                .attr("y", yScale(normMean + (2 * normStdDev)))
                .attr("height", yScale(normMean - (2 * normStdDev)) -
                      yScale(normMean + (2 * normStdDev)));
            
            var normLine = plotG.select(".norm_line")
                .attr("x1", 0).attr("x2", width - margin.left - margin.right)
                .attr("y1", yScale(normMean)).attr("y2", yScale(normMean));

            if (zoomEnabled === true) {
                zoom = d3.behavior.zoom()
                    .x(xScale).y(yScale)
                    .scaleExtent([1,10])
                    .on("zoom", zoomChanged);

                plotG.select(".pointer_mask").call(zoom);
            }

            function zoomChanged() {
                var trans = zoom.translate(),
                scale = zoom.scale();

                var tx = Math.min(0, Math.max(width * (1 - scale), trans[0]));
                var ty = Math.min(0, Math.max(height * (1 - scale), trans[1]));

                zoom.translate([tx, ty]);

                plotSvg.select(".x.axis").call(xAxis);
                plotSvg.select(".y.axis").call(yAxis);

                symbols.each(function() {
                    d3.select(this)
                        .attr("cx", function(d) { return x(d); })
                        .attr("cy", function(d) { return y(d); });
                });
                
                normLine.attr("y1", yScale(normMean)).attr("y2", yScale(normMean));
            }
        });
    }

    function x(d) {
        return xScale(d[0]);
    }
    function y(d) {
        return yScale(d[1]);
    }

    // Layout
    longPlot.margin = function(d) {
        if (!arguments.length) return margin;
        margin = d;
        return longPlot;
    };
    longPlot.height = function(d) {
        if (!arguments.length) return height;
        height = d;
        return longPlot;
    };
    longPlot.width = function(d) {
        if (!arguments.length) return width;
        width = d;
        return longPlot;
    };
    longPlot.minY = function(d) {
        if (!arguments.length) return minY;
        minY = d;
        return longPlot;
    };

    // Data access
    longPlot.x = function(d) {
        if (!arguments.length) return xValue;
        xValue = d;
        return longPlot;
    };
    longPlot.y = function(d) {
        if (!arguments.length) return yValue;
        yValue = d;
        return longPlot;
    };
    longPlot.extra = function(d) {
        if (!arguments.length) return extraValues;
        extraValues = d;
        return longPlot;
    };

    // Labels
    longPlot.title = function(d) {
        if (!arguments.length) return title;
        title = d;
        return longPlot;
    };
    longPlot.xAxisLabel = function(d) {
        if (!arguments.length) return xAxisLabel;
        xAxisLabel = d;
        return longPlot;
    };
    longPlot.yAxisLabel = function(d) {
        if (!arguments.length) return yAxisLabel;
        yAxisLabel = d;
        return longPlot;
    };

    // Set/get for normative data
    longPlot.normStdDev = function(d) {
        if (!arguments.length) return normStdDev;
        normStdDev = d;
        return longPlot;
    };
    longPlot.normMean = function(d) {
        if (!arguments.length) return normMean;
        normMean = d;
        return longPlot;
    };

    // Enable/disable zoom functionality
    longPlot.zoomEnabled = function(d) {
        if (!arguments.length) return zoomEnabled;
        zoomEnabled = d;
        return longPlot;
    };

    // Tooltips
    longPlot.tooltipsEnabled = function(d) {
        if (!arguments.length) return tooltipsEnabled;
        tooltipsEnabled = d;
        return longPlot;
    };
    longPlot.tooltipHtml = function(d) {
        if (!arguments.length) return tooltipHtml;
        tooltipHtml = d;
        return longPlot;
    };

    // On click handler for points
    longPlot.pointOnClickHandler = function(d) {
        if (!arguments.length) return pointOnClickHandler;
        pointOnClickHandler = d;
        return longPlot;
    };

    return longPlot;
}
